/*
 * cdcc.c: dcc client offers. 
 *
 * Written By Colten Edwards <edwac@sasknet.sk.ca> 
 *
 * Copyright(c) 1996
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */


#include "irc.h"

#ifdef ESIX
# include <lan/net_types.h>
#endif /* ESIX */

#if defined(ISC30) && defined(_POSIX_SOURCE)
# undef _POSIX_SOURCE
#include <sys/stat.h>
# define _POSIX_SOURCE
#else
# include <sys/stat.h>
#endif /* ICS30 || _POSIX_SOURCE */

#include <dirent.h>
#include <sys/stat.h>
#include "talkd.h"
#include "server.h"
#include "ircaux.h"
#include "whois.h"
#include "lastlog.h"
#include "ctcp.h"
#include "dcc.h"
#include "hook.h"
#include "vars.h"
#include "window.h"
#include "output.h"
#include "newio.h"
#include "list.h"
#include "timer.h"
#include "screen.h"
#include "pluto.h"

void	dcc_addpack _((char *, char *, char *));
void	dcc_delpack _((char *, char *, char *));
void	dcc_listpack _((char *, char *, char *));
void	dcc_listp _((char *, char *, char *));
void	dcc_offer _((char *, char *, char *));
void	dcc_notice _((char *, char *, char *));
void	dcc_help _((char *, char *, char *));
void	dcc_mhelp _((char *, char *, char *));
void	dcc_showme _((char *, char *, char *));
void	dcc_getpack _((char *, char *, char *));
void	dcc_sendpack _((char *, char *, char *));
void	dcc_show_queue _((char *, char *, char *));
void	dcc_timer _((char *, char *, char *));

extern int dcc_active_count;
extern void dcc_filesend _((char *));
extern void dcc_show_active _((char *));

FileList *QueueList = NULL;

struct 
{
	char * name;
	int user;
	void (*func) _((char *, char *, char *));
} cmdslist[2][12] =
{{
	{ "LIST",	1, dcc_listpack }, /* User Commands */
	{ "PLIST",	0, dcc_offer },
	{ "OFFER",	0, dcc_addpack },
	{ "GET",	0, dcc_getpack },
	{ "SEND",	1, dcc_sendpack },
	{ "RESEND",	1, dcc_sendpack },
	{ "DOFFER",	0, dcc_delpack },
	{ "NOTICE",	0, dcc_notice },
	{ "HELP",	1, dcc_mhelp },
	{ "QUEUE",	0, dcc_show_queue },
	{ "TIMER",	1, dcc_timer },
	{ "",		0, (void (*) _((char *, char *, char *))) NULL},
},{
	{ "LIST",	0, dcc_listp }, /* Local Commands */
	{ "PLIST",	0, dcc_offer },
	{ "OFFER",	0, dcc_addpack },
	{ "GET",	0, dcc_getpack },
	{ "SEND",	0, dcc_sendpack },
	{ "RESEND",	0, dcc_sendpack },
	{ "DOFFER",	0, dcc_delpack },
	{ "NOTICE",	0, dcc_notice },
	{ "HELP",	0, dcc_help },
	{ "QUEUE",	0, dcc_show_queue },
	{ "TIMER",	0, dcc_timer },
	{ "",		0, (void (*) _((char *, char *, char *))) NULL},
}
};

FileList *file_lists = NULL;
FileList *new = NULL;

static char pack_description[1000];
static int  packnumber = 0;
static int dcc_timer_val = 0;
#define DEFAULT_TIMER_VAL 60

int number_in_queue = 0;

char *msgcdcc(char *from, char *to, char *args ) 
{
int i; 
char *p = NULL, *q;
char *ptr;
char *xdcc, *comm;
	ptr = args;
	if ((stristr(args, "XDCC") != NULL) || (stristr(args,"CDCC") != NULL) || (stristr(args, "BDCC") != NULL)) {
		if (file_lists == NULL)
			return ptr;
		malloc_strcpy(&p, args);
		q = p;
		xdcc = next_arg(p, &p);
		comm = next_arg(p, &p);
		for (i = 0; *cmdslist[0][i].name; i++ ) {
			if ((my_stricmp(comm, cmdslist[0][i].name) == 0) && (cmdslist[0][i].func != NULL) && cmdslist[0][i].user){
				(*cmdslist[0][i].func)(p, from, NULL);
				ptr = NULL;
				break;
			}
		}
		if (ptr) {
			send_to_server("NOTICE %s :\002Unknown CDCC command error. Try /msg %s CDCC help\002", from, to);
		}
		new_free(&q);
	}
	return ptr;
}

void cdcc(char *command, char *args, char *subargs ) 
{
int i; char *from, *to;
	from = to = NULL;
	if (!args || !*args) {
		bitchsay("Try /CDCC help");
		return;
	}
	from  = next_arg(args, &args);
	for (i = 0; *cmdslist[1][i].name; i++ ) {
		if ((my_stricmp(from, cmdslist[1][i].name) == 0) && (cmdslist[1][i].func != NULL) && !cmdslist[1][i].user){
			(*cmdslist[1][i].func)(args, from, to);
			break;
		}
	}
}

char *do_bdcc(CtcpEntry *ctcp, char *from, char *to, char *args )
{
char *p;
int i;
int channel_ctcp = 0;
char *ptr = args;
	if (!args || !*args)
		return NULL;
	p  = next_arg(args, &args);
	if (*to == '#')
		channel_ctcp++;
	for (i = 0; *cmdslist[0][i].name; i++ ) {
		if ((my_stricmp(p, cmdslist[0][i].name) == 0) && (cmdslist[0][i].func != NULL) && cmdslist[0][i].user){
			(*cmdslist[0][i].func)(args, from, to); 
			ptr = NULL;
			break;
			
		}
		ptr = NULL;
	}
	return ptr;
}

void dcc_show_queue(char *args, char *from, char *to)
{
FileList *tmp = QueueList;
	if (tmp) {
		put_it(" %-10s %-25s %s", "Nick", "Time", "Description");
		for (tmp = QueueList; tmp; tmp = tmp->next)
			put_it(" %-10s %-25s %s ", tmp->nick, my_ctime((tmp->timequeue)), tmp->description); 
	} else
		bitchsay("No Files Queued for send");
}

FileList *find_lastqueue(void) 
{
FileList *tmp = NULL, *tmp1 = NULL;
	if (number_in_queue) {
		for (tmp = QueueList; tmp; tmp = tmp->next) {
			if (tmp->next == NULL)
				break;
			else
				tmp1 = tmp;
		}
	}
	return tmp1;
}

void dcc_sendfrom_queue(void)
{
FileList *tmp = NULL, *tmp1 = NULL;
char * p = NULL;
	if ((tmp = find_lastqueue()) != NULL) {
		/* more than one in queue */
		tmp1 = tmp->next;
		if (tmp1)
			tmp->next = NULL;
	} else {
		/* only one or possibly none in queue */
		if (QueueList)
			tmp1 = QueueList;
	}
	if (tmp1) {
		bitchsay("Sending file to %s   filename %s", tmp1->nick, tmp1->filename);
		p = (char *) new_malloc(strlen(tmp1->nick)+strlen(tmp1->filename)+4);
		sprintf(p, "%s %s", tmp1->nick, tmp1->filename);
		dcc_filesend(p);
		new_free(&p);
		if (tmp1->nick) new_free(&tmp1->nick);
		if (tmp1->description) new_free(&tmp1->description);
		if (tmp1->filename) new_free(&tmp1->filename);
		new_free((char **)&tmp1);
		if (number_in_queue) number_in_queue--;
	}
}

void dcc_queue(char *nick, char *filename, char *description)
{
FileList *tmp = NULL;
	bitchsay("dcc queue called with nick = %s and  filename = %s", nick, filename);
	tmp = (FileList *) new_malloc(sizeof(FileList));
	memset(tmp, 0, sizeof(FileList));
	malloc_strcpy(&tmp->nick, nick);
	malloc_strcpy(&tmp->filename, filename);
	malloc_strcpy(&tmp->description, description);
	tmp->timequeue = time(NULL);
	number_in_queue++;
	send_to_server("NOTICE %s :\002File Sends are currently busy. File %s has been Queued\002", nick, description); 
	send_to_server("NOTICE %s :I will send as soon as a spot on the queue is open.", nick); 
	add_to_list((List **)&QueueList, (List *)tmp);	
}


void dcc_sendpack (char *args, char *from, char *to)
{
static int lame_counter = 0;
char *channel, *p, *nick = NULL;
int file_no = 0;
FileList *files = file_lists;

	p = next_arg(args, &args);
	if (p && *p && *p =='#')
	{
		p++;
		file_no = atoi(p);
		if (file_no  <= 0)
			return;
	} 
	if ((nick = next_arg(args, &args)) == NULL)
		nick = from;
	for (files = file_lists; files != NULL; files = files->next) {
		if (files->packnumber == file_no) {
			files->gets++;
			break;
		}
	}
	channel = get_channel_by_refnum(0);
	if (files && get_int_var(AUTOCK_VAR) && (channel != NULL)) {
		if (you_is_opped(channel))
			send_to_server("KICK %s %s :\002LaMe CTCP KiCk CoUnT #%d\002", channel, from, ++lame_counter);
		return;
	}
	if (files != NULL) {
/*
		if (get_int_var(DCC_SEND_LIMIT_VAR) && (dcc_active_count > get_int_var(DCC_SEND_LIMIT_VAR))) {
			dcc_queue(nick, files->filename, files->description);
			return;
		}
*/
		p = (char *) new_malloc(strlen(nick)+strlen(files->filename)+4);
		sprintf(p, "%s %s", nick, files->filename);
		dcc_filesend(p);
		dcc_show_active(NULL);
		new_free(&p);
		return;
	}
}

void dcc_getpack (char *args, char *from, char *to)
{
/*	nick = next_arg(args, &args);*/
}

void dcc_listpack (char *args, char *from, char *to) 
{
FileList *files = file_lists;
	if (files != NULL) {
		send_to_server("NOTICE %s :Files Offered: /ctcp %s CDCC send #N for pack N", from, nickname);
		if (get_int_var(DCC_SEND_LIMIT_VAR))
			send_to_server("NOTICE %s :    [%d packs %d/%d slots open]", from, 
				packnumber, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR));
		else
			send_to_server("NOTICE %s :    [%d packs]", from, packnumber);
		for (files = file_lists; files != NULL; files = files->next) {
			send_to_server("NOTICE %s :\002#%d\002:  %s [\002%d\002 File%s/\002%-2.3f\002 Kb] %d gets", from, 
			files->packnumber, files->description, 
			files->numberfiles, files->numberfiles == 1 ? "":"s",
			files->filesize/(double)1000, files->gets);
		}	
	} 
	return;
}

void dcc_listp (char *args, char *from, char *to) 
{
FileList *files = file_lists;
	if (files != NULL) {
		for (files = file_lists; files != NULL; files = files->next) {
			put_it(":\002#%d\002:  %s [\002%d\002 File%s/\002%-2.3f\002 Kb] %d gets",
			files->packnumber, files->description, 
			files->numberfiles, files->numberfiles == 1 ? "":"s",
			files->filesize/(double)1000, files->gets);
		}	
	} else put_it("%s :%s", from, "No Files Offered");
	return;
}

void add_desc(char *args, char *subargs)
{
FileList *files = NULL;
int renumber;
	strcpy(pack_description, subargs);
	if (new) {
		malloc_strcpy(&new->description, pack_description);
		add_to_list((List **) &file_lists, (List *)new);
		bitchsay("Created new pack : [\002%d\002 File%s/%-2.3f Kb]", new->numberfiles,
			new->numberfiles == 1? "":"s", new->filesize/(double)1000);
		for (renumber = 1, files = file_lists; files != NULL; files= files->next, renumber++)
			files->packnumber = renumber;
	}
}

void add_file(char *args, char *subargs) 
{
char path[BIG_BUFFER_SIZE+1];
char filebuf[BIG_BUFFER_SIZE+1];
char *thefile, *expand;
DIR *dp;
struct dirent *dir;
struct stat stbuf;
static int been_here = 0;

	if (been_here == 0) {
		been_here ++;
		file_lists = NULL;
	}
	if (new) new = NULL;	
	strcpy(path, subargs);
        if ((thefile = strrchr(path, '/')) != NULL)
              *thefile++ = '\0';
        else {
              strcpy(path, "~");
              thefile = subargs;
        }
        expand = expand_twiddle(path);
	dp = opendir(expand);
	if (!dp) {
		bitchsay("You cannot access dir %s", expand);
		new_free(&expand);
		return;
	}
	bitchsay("Searching %s/%s", expand, thefile);
        new = (FileList *)new_malloc(sizeof(FileList));
        memset(new, 0, sizeof(FileList));
	while ((dir = readdir(dp)) != NULL) {
                if (dir->d_ino == 0)
                       continue;
                if (!do_rmatch(dir->d_name, thefile))
                       continue;
                sprintf(filebuf, "%s/%s", expand, dir->d_name);
                stat(filebuf, &stbuf);
                if (S_ISDIR(stbuf.st_mode))
                       continue;
                new->filesize += stbuf.st_size;
                new->numberfiles++;
	}
	if (new->numberfiles) {
                new->packnumber = ++packnumber;
		sprintf(filebuf, "%s/%s", expand, thefile);
		malloc_strcpy(&new->filename, filebuf);
		bitchsay("Added %d files to Pack....", new->numberfiles);
		add_wait_prompt("Description of Pack? ", add_desc, "", WAIT_PROMPT_LINE);
	} else { 
		bitchsay("No Files for Pack found, Aborting....");
		if (new) new_free((char **)&new);
		new = NULL;
	}
	new_free(&expand);
}

void dcc_addpack (char *args, char *from, char *to)  
{
	add_wait_prompt("Add Files to Pack? ", add_file, "", WAIT_PROMPT_LINE);
}

void clear_offer(FileList **clientlist)
{
        FileList        *Client, *tmp;
        for (Client = *clientlist; Client;)
        {
              new_free(&Client->filename);
              new_free(&Client->description);
              new_free(&Client->nick);
              tmp = Client->next;
              new_free((char **)&Client);
              Client = tmp;
         }
         *clientlist = NULL;
}

void del_file(char *args, char *subargs) 
{
FileList *files = file_lists;
FileList *delpack = NULL;
char *p;
int packno;
int renumber;

	if (*subargs != '#' && *subargs != '*') {
		say("Need to specify #0 or * for all");
		return;
	}
	p = subargs;
	if (*p != '*')
	{
		p++;
		packno = atoi(p);
		if (!packno) {
			say("Need to number between 1 and last pack number");
			return;
		}
	} 
	else
	{
		clear_offer(&file_lists);
		packnumber = 0;
		return;
	}
	if (files != NULL) {
		char *tmpname = NULL;
		for (files = file_lists; files != NULL; files = files->next) {
			if (files->packnumber == packno) {
				tmpname = files->filename;
				break;
			}
		}
		if (tmpname && (delpack = (FileList *)remove_from_list((List **)&file_lists, tmpname)) != NULL) {
			new_free(&delpack->filename);
			new_free(&delpack->description);
			new_free((char **)&delpack);
			delpack = NULL;
		}
		for (packnumber = 0, renumber = 1, files = file_lists; files != NULL; files = files->next, renumber++, packnumber++) 
			files->packnumber = renumber;
	} else say("No Packs to delete");
}

void dcc_delpack (char *args, char *from, char *to)
{
	add_wait_prompt("Delete Pack [* for all]? ", del_file, "", WAIT_PROMPT_LINE);
}

void dcc_offer (char *args, char *from, char *to)
{
FileList *files = file_lists;
char *channel = NULL;
char what[BIG_BUFFER_SIZE + 1];
	if (files == NULL)
	{
		bitchsay("No Files Offered");
		return;
	}
	if (get_channel_by_refnum(0)) {
		malloc_strcpy(&channel, get_channel_by_refnum(0));
		sprintf(what, "Files Offered: /Ctcp %s \002CDCC\002 list", nickname);
		send_text(channel, what, NULL);
		if (get_int_var(DCC_SEND_LIMIT_VAR))
			sprintf(what, "    [%d packs %d/%d slots open]",packnumber, 
				get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR));
		else
			sprintf(what, "    [%d packs]",packnumber); 
		send_text(channel, what, NULL);

		for (files = file_lists; files != NULL; files = files->next) 
		{
			sprintf(what, "\002#%d\002:  %s [\002%d\002 File%s/\002%-2.3f\002 Kb] %d gets",
			files->packnumber, files->description ? files->description: "No Description", 
			files->numberfiles, files->numberfiles == 1 ? "":"s",
			files->filesize/(double)1000, files->gets);
			send_text(channel,what, NULL);
		}	
		new_free(&channel);
	}
}

void dcc_notice (char *args, char *from, char *to)
{
char *channel = NULL;
char *msg = NULL;

	if (get_channel_by_refnum(0)) {
		if (file_lists)
		{
			malloc_strcpy(&channel, get_channel_by_refnum(0));
			malloc_sprintf(&msg, "\002== FILES Offered\002  /Ctcp %s \002CDCC\002 list", nickname);
			send_text(channel, msg, NULL);
			if (get_int_var(DCC_SEND_LIMIT_VAR))
				malloc_sprintf(&msg, "    [%d packs %d/%d slots open]",
					packnumber, get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR));
			else
				malloc_sprintf(&msg, "    [%d packs]",packnumber);
			
			send_text(channel, msg, NULL);	
			new_free(&channel);
			new_free(&msg);
		}
		else
			bitchsay("No Files to offer");
	}
}

int do_plist_timer(void *args)
{
	char *p = (char *)args;
	
	char *timer, *channel, *q = NULL;
	char what[BIG_BUFFER_SIZE+1];
	FileList *files = file_lists;
	
	malloc_strcpy(&q, p);
	timer = next_arg(q, &q);
	channel = next_arg(q, &q);	

	if (dcc_timer_val == 0 || !im_on_channel(channel) || !files)
	{
		new_free(&p);
		new_free(&q);
		return 0;
	}
	
	add_timer(-1, dcc_timer_val, do_plist_timer, (char *)args, NULL);
	sprintf(what, "Files Offered: /Ctcp %s \002CDCC\002 list", nickname);
	send_text(channel, what, NULL);
	if (get_int_var(DCC_SEND_LIMIT_VAR))
		sprintf(what, "    [%d packs %d/%d slots open]",packnumber, 
			get_int_var(DCC_SEND_LIMIT_VAR)-dcc_active_count, get_int_var(DCC_SEND_LIMIT_VAR));
	else
		sprintf(what, "    [%d packs]",packnumber); 
	send_text(channel, what, NULL);

	for (files = file_lists; files != NULL; files = files->next) 
	{
		sprintf(what, "\002#%d\002:  %s [\002%d\002 File%s/\002%-2.3f\002 Kb] %d gets",
		files->packnumber, files->description ? files->description: "No Description", 
		files->numberfiles, files->numberfiles == 1 ? "":"s",
		files->filesize/(double)1000, files->gets);
		send_text(channel,what, NULL);
	}	
	new_free(&timer);
	return 0;
}

void add_plist_timer(char *args, char *subargs)
{
char *p = NULL;
char *timer, *channel;
	if (!subargs && !*subargs)
	{
		dcc_timer_val = 0;
		return;
	}
	if (*subargs == '-')
	{
		bitchsay("Removing timer plist");
		dcc_timer_val = 0;
		return;
	}
	timer = next_arg(subargs, &subargs);
	dcc_timer_val = atoi(timer);
	if (dcc_timer_val < DEFAULT_TIMER_VAL)
		dcc_timer_val = DEFAULT_TIMER_VAL;

	if (args && *args)
		channel = next_arg(subargs, &subargs);
	else
		channel = get_channel_by_refnum(0);

	if (!channel)
	{
		bitchsay("No Channel");
		return;		
	}
	bitchsay("Timer set to %d secs for %s", dcc_timer_val, channel);
	malloc_sprintf(&p, "%d %s", dcc_timer_val, channel);
	add_timer(-1, dcc_timer_val, do_plist_timer, p, NULL);
}

void dcc_timer(char *args, char *from, char *to)
{
	add_wait_prompt("Delay for plist? [- to delete]", add_plist_timer, "", WAIT_PROMPT_LINE);
}

void dcc_help (char *args, char *from, char *to)
{
	put_it("\002CDCC\002: get	- pack from a user");
	put_it("\002CDCC\002: send 	- send a pack to nick");
	put_it("\002CDCC\002: offer	- add a pack");
	put_it("\002CDCC\002: doffer	- delete a pack");
	put_it("\002CDCC\002: list	- list pack for you");
	put_it("\002CDCC\002: plist	- public list a pack");
	put_it("\002CDCC\002: notice	- announce to channel a pack");
	put_it("\002CDCC\002: queue	- show queue");
	put_it("\002CDCC\002: timer	- plist timer");
}

void dcc_mhelp (char *args, char *from, char *to)
{
	send_to_server("NOTICE %s :\002 SEND #  - send a file pack", from);
	send_to_server("NOTICE %s :\002 LIST    - list file packs", from);
	send_to_server("NOTICE %s :\002 HELP    - This help msg", from);
}
